;**************************************************************************
;                               I2CLOW.INC
;
;			Autor: Mikel Etxebarria
;                    (c) Microsystems Engineering (Bilbao)
;
;Conjunto de rutinas de bajo nivel que permiten la transmisin/recepcin
;a nivel de byte entre dispositivos I2C. Igualmente generan las condiciones
;de inicio y de stop as como detectar/generar los bits ACK y NACK. Estas
;rutinas estn pensadas para ser utilizadas con elementos de la gama media
;16cXX
;
;                       
;***************************************************************************
;Definicin de registros empleados por el software de interface al bus I2C 
;con los dispositivos de la gama media 16cxx
;

Clkout          equ     CLKIN >> 2		 ;Frecuencia de trabajo (Clkin/4)

_4uS_Delay      set     (Clkout/.250000) ;Constante para 4 uS
_47uS_Delay     set     (Clkout/.212766) ;Constante para 4.7 uS
_5uS_Delay      set     (Clkout/.200000) ;Constante para 5 uS


                CBLOCK  I2c_var ;Reserva de memoria a partir de I2c_var que
                		;ser la primera libre disponible
		DataByte        ;Dato que va a ser transmitido
                BitCount        ;N de bits a transmitir/recibir (0:7)
                Bus_Status      ;Registro de estado del bus I2C
                Bus_Control     ;Registro de control del bus I2C
                DelayCount      ;Variable de temporizacin
                DataByteCopy    ;Copia del dato a transmitir
                ENDC            ;Fin de la reserva

;***************************************************************************
;Definicin de las lneas del bus I2C
;***************************************************************************

#define _SCL    	PORTB,7
#define _SDA		PORTB,6
#define _SCL_TRIS       PORTB,7
#define _SDA_TRIS       PORTB,6
#define _WRITE_         0
#define _READ_          1

;***************************************************************************
;Definicin de los bits del registro de estado del bus I2C
;***************************************************************************

#define _Bus_Busy       Bus_Status,0    ;Bus I2C ocupado
#define _Abort          Bus_Status,1    ;Comunicacin abortada
#define _Txmt_Progreso  Bus_Status,2    ;Transmisin en progreso
#define _Rcv_Progreso   Bus_Status,3    ;Recepcin en progreso
#define _Txmt_Acabada   Bus_Status,4    ;Transmisin finalizada
#define _Rcv_Acabada    Bus_Status,5    ;Recepcin finalizada
#define _Fatal_Error    Bus_Status,6    ;Error en comunicacin
#define _ACK_Error      Bus_Status,7    ;Error en bit ACK

;***************************************************************************
;Definicin de los bits del registro de control del bus I2C
;***************************************************************************

#define _Ultimo_Byte    Bus_Control,2   ;Indica si es el ltimo byte a recibir


;**************************************************************************
;               Macro de propsito general
;**************************************************************************

LIBERA_BUS      MACRO
                bsf	STATUS,RP0	;Selecciona banco 1                
                bsf     _SDA	        ;SDA en triestado
                bsf     _SCL            ;SCL en triestado
                bcf	STATUS,RP0	;Selecciona banco 0 de dato
		bcf	_Bus_Busy
                ENDM

;**************************************************************************
;                               Inicia_Bus
;
;Inicializacin del Bus I2C
;**************************************************************************

Inicia_Bus:

                bcf     STATUS,RP0      ;Selecciona pgina 0 de datos
                movf    PORTB,W         ;Carga puerta B
                andlw   0x3f
                movwf   PORTB           ;Pone a 0 los latch internos
                                        ;de las lneas SCL y SDA
                LIBERA_BUS              ;Coloca SDA y SCL en triestado
                clrf    Bus_Control     ;Borra registro de control
                clrf    Bus_Status      ;Borra registro de estado                
		return

;**************************************************************************
;                       Txt_Start_Bit
;
;Transmisin de la condicin de inicio
;**************************************************************************

Txt_Start_Bit:

		bsf     STATUS,RP0      ;Selecciona pgina 1
                bsf     _SDA            ;Pone como entradas las seales SDA y
                bsf     _SCL            ;SCL que al estar con resistencia
                            	        ;pull-up, las lneas SDA y SCL se man-
                                	;tienen a "1"
                call 	Delay4uS
                bcf     _SDA            ;La seal SDA se configura como salida
                                        ;que saca el "0" del latch interno ha-
                                        ;ciendo que la lnea SDA caiga a "0"
                                        ;mientras SCL se mantiene a "1" (con-
                                        ;dicin de inicio)
                call 	Delay47uS
		bcf	STATUS,RP0	;Selecciona banco 0 de dato
                bsf 	_Bus_Busy       ;Activa flag de bus ocupado		
                return

;**************************************************************************
;                               TxT_Stop_Bit
;
;Transmisin de la condicin de stop
;**************************************************************************

Txt_Stop_Bit:
                bsf     STATUS,RP0      ;Selecciona pgina 1 de datos
                bcf     _SCL            ;Las seales SCL y SDA se programan
                bcf     _SDA            ;como salidas. El contenido de los
                                        ;latch internos hace que las lneas
                                        ;SCL y SDA se pongan a "0"

                bsf     _SCL            ;Esta seal se programa ahora como
                                        ;entrada en alta impedancia que, gra-
                                        ;cias a las pull-up internas, pone a
                                        ;"1" la lnea SCL.

                call 	Delay4uS

                bsf     _SDA            ;Esta seal se programa ahora como
                                        ;entrada en alta impedancia que, gra-
                                        ;cias a las pull-up internas, genera un
                                        ;flanco de subida en la lnea SDA (con-
                                        ;dicin de stop).

                call 	Delay47uS       ;Asegura que no se enva una nueva
                                      	;condicin de start justo despus de la
                                       	;de stop.
		bcf	STATUS,RP0	;Selecciona banco 0 de dato
                bcf     _Bus_Busy       ;Tras el stop se considera que el bus
                                        ;est libre, borrando el bit de estado
                return


;**************************************************************************
;                               Txt_Abort
;
;Abortar la transmisin
;**************************************************************************

Txt_Abort:

                call    Txt_Stop_Bit    ;Transmite la condicin de stop
                bsf     _Abort          ;Activa flag en registro de estado
                return

;**************************************************************************
;                               Txt_Byte
;
;Transmisin de un byte de datos
;El Byte a transmitir debe cargarse previamente en el registro DataByte
;**************************************************************************

Txt_Byte:

                movf    DataByte,w      ;Realiza una copia del byte a trans-
                movwf   DataByteCopy    ;mitir (pos.0x0e)
                bsf     _Txmt_Progreso  ;Activa flag de transmisin en curso
                bcf     _Txmt_Acabada   ;Desactiva flag de finalizacin
                movlw   0x08
                movwf   BitCount        ;Nmero de bits a transmitir
                bsf     STATUS,RP0      ;Selecciona pgina 1 de datos

TxtNextBit:     clrwdt                  ;Refresco del WDT
                bcf     _SCL            ;Pone a "0" la lnea SCL
                rlf     DataByteCopy,f  ;Desplaza bit de ms peso
                bcf     _SDA            ;Pone a "0" la lnea SDA
                btfsc   STATUS,C        ;Testea bit a sacar sito en el carry
                bsf     _SDA            ;Saca "1" por la lnea SDA poniendo
                                        ;la seal SDA como entrada pull-up
                call    Delay47uS
                bsf     _SCL            ;Pone a "1" la lnea SCL poniendo
                                        ;la seal SCL como entrada pull-up
                call    Delay4uS
                decfsz  BitCount,f      ;Decrementa contador de bits
                goto    TxtNextBit

;Chequeo del bit ACK

                bcf     _SCL            ;Pone a "0" la lnea SCL
                bsf     _SDA            ;Pone a "1" la lnea SDA para detec-
                                        ;tar cundo el slave la pone a "0"
                call    Delay47uS
                bsf     _SCL            ;Pone a "1" lnea de reloj
                call    Delay4uS

                bcf     STATUS,RP0      ;Selecciona pgina 0 de datos
                btfsc   _SDA            ;La seal SDA debe estar a "0" envia-
                                        ;do por el slave
                goto    TxtErrorAck
                bsf     STATUS,RP0      ;Selecciona pgina 1 de datos
                bcf     _SCL            ;Pone a "0" la lnea SCL

                bcf	STATUS,RP0	;Selecciona banco 0 de dato
		bcf     _Txmt_Progreso  ;Borra flag de transmisin en curso
                bsf     _Txmt_Acabada   ;Activa flag de fin de transmisin
                bcf     _ACK_Error      ;Bit ACK okey
                bcf     _Bus_Busy       ;Flag de que el bus est libre
                return

TxtErrorAck:    LIBERA_BUS              
                bcf     _Txmt_Progreso  ;Borra flag de transmisin en curso
                bcf     _Txmt_Acabada   ;Borra flag de fin de transmisin
                bsf     _ACK_Error      ;Activa bit de error ACK (NACK)
                return

;***************************************************************************
;                               RcvByte
;
;Recepcin de un byte desde el slave
;El dato recibido se almacena en el registro DataByte.
;El bit _Ultimo_Byte del registro de control determina si se espera recibir
;ms bytes desde el slave, en cuyo caso se le manda ACK o, si es el ltimo,
;en cuyo caso se enva NACK.
;***************************************************************************

RcvByte:
                bsf     _Rcv_Progreso   ;Flag de recepcin en curso
                bcf     _Rcv_Acabada    ;Desactiva flag de fin de recepcin
                movlw   0x08            ;Carga el contador de bits a recibir
                movwf   BitCount

RcvNextBit:
                clrwdt                  ;Refresco del WDT
                bsf     STATUS,RP0      ;Seleccin de pgina 1 de datos
                bcf     _SCL            ;Pone a "0" la lnea de reloj
                bsf     _SDA            ;Pone a "1" la lnea de datos
                call    Delay47uS
                bsf     _SCL            ;El reloj pasa a "1". El bit de datos
                                        ;es puesto por el slave
                call    Delay4uS

                bcf     STATUS,RP0      ;Seleciona pgina 0 de datos
                bcf     STATUS,C	;Pone el flag carry a "0"
                btfsc   _SDA            ;Chequea el nivel en lnea de datos
                bsf     STATUS,C	;Est a "1", activa el carry
                rlf     DataByte,f      ;Desplaza a la izda. (primero el bit
                                        ;de ms peso)
                decfsz  BitCount,f      ;Decrementa contador de bits a recibir
                goto    RcvNextBit

;El Master genera ACK si no es el ltimo byte a recibir. Si es el ltimo
;se genera NACK y a continuacin se debe generar la condicin de stop.

                bsf     STATUS,RP0      ;Selecciona pgina 1 de datos
                bcf     _SCL            ;Pone a "0" la lnea de reloj
                bcf     _SDA            ;Pone a "0" la lnea de datos (ACK)
                btfsc   _Ultimo_Byte    ;Determina si es el ltimo byte
                bsf     _SDA            ;Pone a "1" la lnea de datos (NACK)
                call    Delay47uS
                bsf     _SCL            ;Pone a "1" la lnea de reloj
                call    Delay4uS
                bcf     _SCL            ;Pone a "0" la lnea de reloj

                bcf	STATUS,RP0	;Selecciona banco 0 de dato
		bcf     _Rcv_Progreso   ;Desactiva flag de recepcin en curso
                bsf     _Rcv_Acabada    ;Activa flag de fin de recepcin
                bcf     _ACK_Error      ; ACK ok
                return

;****************************************************************************
;Rutinas de temporizacin de propsito general
;****************************************************************************
;
Delay5uS:
                movlw   ((_5uS_Delay-5)/3+1)
delay:
                movwf   DelayCount
                decfsz  DelayCount,f
                goto 	$-1
                return

Delay47uS:      movlw   ((_47uS_Delay)-8/3+1)
                goto 	delay

Delay4uS:       movlw   ((_4uS_Delay)-8/3+1)
                goto 	delay

