;****************************************************************************************
;					TECLADO.INC
;
;				Autor: Mikel Etxebarria
;			(c) Microsystems Engineering (Bilbao)
;
;Presenta un conjuntos de rutinas que permiten la gestin de un teclado hexadecimal organi-
;zado en una matriz de 4 x 4 tal y como es el caso del teclado incorporado en la tarjeta
;de ampliacin MicroPic Trainer PLUS de Microsystems Engineering.
;
;Este fichero se debe incluir en los futuros programas fuente mediante la directiva INCLUDE.
;
;Key_Scan: 	realiza un barrido del teclado y detecta si hay alguna tecla pulsada. La
;		variable "Tecla" retorna con el cdigo de la tecla pulsada o el 0x80 en caso 
;		de no haber ninguna.
;
;Key_BCD:	Convierte el cdigo de tecla en cdigo BCD (del 0 a F). Antes de llamar a la
;		rutina la variable "Tecla" contiene el cdigo de tecla. Al finalizar, la
;		rutina devuelve el cdigo BCD en la misma variable "Tecla".
;
;BCD_7seg:	Convierte el cdigo BCD en cdigo 7 segmentos. A la llamada de la rutina, el
;		W debe contener el cdigo BCD y al finalizar la misma W retorna con el c-
;		digo 7 segmentos.
;
;El teclado hex. se supone conectado a la puerta B y su disposicin es la siguiente:
;
;		 RB0 RB1 RB2 RB3
;		  ^   ^   ^   ^	      TECLA	     CODIGO
;                 |   |   |   |       -----	--------------
;		|---|---|---|---|	0	01111101 = 0x7D
;	RB4 -->	| 1 | 2 | 3 | F |	1	11101110 = 0xEE
;		|---|---|---|---|	2	11101101 = 0xED
;	RB5 -->	| 4 | 5 | 6 | E |	3	11101011 = 0xEB
;		|---|---|---|---|	4	11011110 = 0xDE
;	RB6 -->	| 7 | 8 | 9 | D |	5	11011101 = 0xDD
;		|---|---|---|---|	6	11011011 = 0xDB
;	RB7 -->	| A | 0 | B | C |	7	10111110 = 0xBE
;		|---|---|---|---|	8	10111101 = 0xBD
;				        9	10111011 = 0xBB
;				        A	01111110 = 0x7E
;				        B	01111011 = 0x7B
;					C	01110111 = 0x77
;					D	10110111 = 0xB7
;					E	11010111 = 0xD7
;					F	11100111 = 0xE7
;
;Hubiera sido ms interesante que RB4-RB7 actuaran como entradas para poder hacer
;uso de la capacidad que tiene el PIC de provocar interrupcin al cambiar de estado
;estas lneas. Ello no es posible dado que, en la Trainer, las lneas RB6 y RB7 forman
;parte tambin del circuito de grabacin.

		cblock	Key_var   	;Inicio de las variables
		Tecla			;Retorno del cdigo de tecla
		Key_1			;N de columnas a explorar
		Key_2               	;Temporal de cdigo
		Key_Delay_1		;Variable de temporizacin
		Key_Delay_2		;Variable de temporizacin
		TRISB_Temp		;Estado temporal de TRISB
		endc						

;**************************************************************************************
;Key_Scan: Rutina de exploracin del teclado. La variable "Tecla" retorna con el cdigo
;de la tecla pulsada o el cdigo 0x80 si no se pulsa ninguna.

Key_Scan:	bsf	STATUS,RP0	;Seleciona pgina 1
		movf	TRISB,W
		movwf	TRISB_Temp	;Salva estado actual de TRISB
		movlw	b'00001111'
		movwf	TRISB		;RB0-RB3 entradas, RB4-RB7 salidas
		bcf	STATUS,RP0	;Selecciona pgina 0
		movlw	4
		movwf	Key_1		;N de columnas a explorar
		movlw	b'01111111'
		movwf	Tecla		;Columna a activar
Key_Scan_1:	movf    Tecla,W
		movwf	PORTB		;Activa columna				
		nop
		movf	PORTB,W
		movwf	Key_2
		subwf	Tecla,W		;Lee las filas
		btfss	STATUS,Z	;Hay alguna pulsada ?
		goto	Key_Scan_2	;Si, hay alguna pulsada
		bsf	STATUS,C	;No hay ninguna en esa columna
		rrf	Tecla,F		;Seleciona siguiente columna
		decfsz	Key_1,F
		goto	Key_Scan_1
		movlw	0x80
		movwf	Tecla		;Retorna cdigo 0x80 (no hay pulsacin)
		movf	TRISB_Temp,W
		bsf	STATUS,RP0	;Selecciona pgina 1
		movwf	TRISB		;Repone TRISB al valor original
		bcf	STATUS,RP0	;Selecciona pgina 0
		return			;Fin de exploracin
				
Key_Scan_2:	movlw	.10		;Bucle de temporizacin de unos 20 mS
		movwf	Key_Delay_1	;para evitar el rebote de los pulsado-
Key_Scan_3:	clrf	Key_Delay_2	;res.
		clrwdt
Key_Scan_4:	decfsz	Key_Delay_2,F
		goto	Key_Scan_4
		decfsz	Key_Delay_1,F
		goto	Key_Scan_3
				
		movf	Tecla,W		;Trs la temporizacin se lee nuevamente
		movwf	PORTB		;si la tecla es la misma. As se evitan
		nop			;los rebotes.
		movf	PORTB,W
		subwf	Key_2,W
		btfss	STATUS,Z	;Es la misma ??
		goto	Key_Scan_1	;No, seguir con la exploracin
		movf	Key_2,W		;Si guardar en variable de salida "Tecla"
		movwf	Tecla
		movf	TRISB_Temp,W
		bsf	STATUS,RP0	;Selecciona pgina 1
		movwf	TRISB		;Repone TRISB al valor original
		bcf	STATUS,RP0	;Selecciona pgina 0
		return			;Fin de exploracin															

;*************************************************************************************
;Key_BCD: Convierte un cdigo el cdigo de tecla que haya en la variable "Tecla" y lo
;convierte a Hex. (0-F). El resultado se devuelve en la variable "Tecla"
 
Key_Tabla	movf	Key_1,W
		addwf	PCL,F		;Calcula desplazamiento
		retlw	0x7d		;0
		retlw	0xee		;1
		retlw	0xed		;2
		retlw	0xeb		;3
		retlw	0xde		;4
		retlw	0xdd		;5
		retlw	0xdb		;6
		retlw	0xbe		;7
		retlw	0xbd		;8
		retlw	0xbb		;9
		retlw	0x7e		;A
		retlw	0x7b		;B
		retlw	0x77		;C
		retlw	0xb7		;D
		retlw	0xd7		;E
		retlw	0xe7		;F
				
Key_BCD:	movf	Tecla,W
 		movwf	Key_2		;Almacena el cdigo temporalmente
 		clrf	Key_1		;Contador BCD a 0
Key_BCD_2	call	Key_Tabla	;Busca cdigo en la tabla
 		subwf	Key_2,W		;Compara con el de la tecla
 		btfsc	STATUS,Z	;Coincide ??
 		goto	Key_BCD_1	;Si
 		incf	Key_1,F		;No, incrementa contador BCD
 		goto	Key_BCD_2
Key_BCD_1	movf	Key_1,W
		movwf	Tecla		;Carga contador BCD en la variable de salida
		return
				 						
;***************************************************************************************
;BCD_7seg:	Rutina de conversin de BCD (0-F) a 7 segmentos. El acumulador W contiene,
;durante la llamada, el cdigo BCD y, al retornar, el correspondiente a 7 segmentos
;
BCD_7seg:	andlw	b'00001111'
		addwf	PCL,F		;Desplazamiento sobre la tabla
		retlw	b'00111111'	;0
		retlw	b'00000110'	;1
		retlw	b'01011011'	;2
		retlw	b'01001111'	;3
		retlw	b'01100110'	;4
		retlw	b'01101101'	;5
		retlw	b'01111101'	;6
		retlw	b'00000111'	;7
		retlw	b'01111111'	;8
		retlw	b'01100111'	;9
		retlw	b'01110111'	;A
		retlw	b'01111100'	;B
		retlw	b'00111001'	;C
		retlw	b'01011110'	;D
		retlw	b'01111001'	;E
		retlw	b'01110001'	;F
				
				
